import logging

logger = logging.getLogger('OPEN_LISA_SDK')
logger.setLevel(logging.DEBUG)

consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logging.Formatter('[%(name)s] %(asctime)s %(levelname)-8s %(message)s'))
logger.addHandler(consoleHandler)

def set_level(level):
  logger.setLevel(level)

def debug(msg):
  logger.debug(msg)

def info(msg):
  logger.info(msg)

def warn(msg):
  logger.warn(msg)

def warning(msg):
  logger.warning(msg)

def error(msg):
  logger.error(msg)


import socket
import serial
from ..domain.exceptions.could_not_connect_to_server import CouldNotConnectToServerException
from ..domain.instruments.instrument import Instrument
from ..domain.protocol.client_protocol import ClientProtocol
from ..logging import log
from ..common.protocol.message_protocol_tcp import MessageProtocolTCP
from ..common.protocol.message_protocol_rs232 import MessageProtocolRS232

class ApiClient:
  def __init__(self, default_string_response_conversion, default_bytearray_response_conversion):
    self._default_string_response_conversion = default_string_response_conversion
    self._default_bytearray_response_conversion = default_bytearray_response_conversion

  def connect_through_TCP(self, host, port):
    try:
      server_address = (host, port)
      sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      sock.connect(server_address)
      self._client_protocol = ClientProtocol(MessageProtocolTCP(sock))
      self._socket_connection = sock
    except Exception as e:
      log.error(e)
      raise CouldNotConnectToServerException("could not connect with server at {} through TCP".format(server_address))

  def connect_through_RS232(self, baudrate):
    # Discover server RS232
    MAX_COM_TO_TRY = 10
    TIMEOUT_TO_WAIT_HANDSHAKE_RESPONSE = 3
    RS232_HANDSHAKE_CLIENT_REQUEST = 'OPEN'
    RS232_HANDSHAKE_SERVER_RESPONSE = 'LISA'

    connection = None
    for i in range(1, MAX_COM_TO_TRY):
      try:
        endpoint = "COM{}".format(i)
        connection = serial.Serial(port=endpoint, baudrate=baudrate, timeout=TIMEOUT_TO_WAIT_HANDSHAKE_RESPONSE)
        MAX_UNSIGNED_INT = 4_294_967_295
        connection.set_buffer_size(rx_size = MAX_UNSIGNED_INT, tx_size = MAX_UNSIGNED_INT)
        if not connection.isOpen():
          connection.open()

        # custom handshake
        connection.write(RS232_HANDSHAKE_CLIENT_REQUEST.encode())
        response = connection.read(len(RS232_HANDSHAKE_SERVER_RESPONSE))
        if len(response) > 0 and str(response.decode()) == RS232_HANDSHAKE_SERVER_RESPONSE:
          log.debug('Detect Open LISA server at {} with baudrate {}'.format(endpoint, baudrate))
          break
        else:
          connection = None
          log.debug("no answer detected from {}".format(endpoint))
      except serial.SerialException as ex:
        log.info('serial exception {}'.format(ex))
        log.debug("could not connect to {}".format(endpoint))
        connection = None

    if not connection:
      raise CouldNotConnectToServerException("could not detect Open LISA server listening through RS232")

    self._client_protocol = ClientProtocol(MessageProtocolRS232(rs232_connection=connection))

  def disconnect(self):
    self._client_protocol.disconnect()
    if (self._socket_connection):
      try:
        self._socket_connection.shutdown(socket.SHUT_RDWR)
      except (socket.error, OSError, ValueError):
        pass
      self._socket_connection.close()


  def get_instruments(self):
    d = self._client_protocol.get_instruments()
    return [Instrument.from_dict(i, self._client_protocol, self._default_string_response_conversion, self._default_bytearray_response_conversion) for i in d]


class MessageProtocol:
    def __del__(self, connection):
        raise NotImplementedError("Please Implement this method: destructor")

    def send_msg(self, msg, encode=True):
        raise NotImplementedError("Please Implement this method: send_msg")

    def receive_msg(self, decode=True):
        raise NotImplementedError("Please Implement this method: receive_msg")

    def __recvall(self, n):
        raise NotImplementedError("Please Implement this method: __recvall")



from .sdk_exception import OpenLISAException

DEFAULT_MESSAGE = "instrument not exists"


class InstrumentNotFoundException(OpenLISAException):
    """
    Raised when the requested instrument not exists
  """

    def __init__(self, message=DEFAULT_MESSAGE):
        self.message = message
        super().__init__(self.message)



from .sdk_exception import OpenLISAException

DEFAULT_MESSAGE = "invalid command"

class InvalidCommandException(OpenLISAException):
  """
    Raised when the SDK is initiated and connection with server could not be established
  """
  def __init__(self, message=DEFAULT_MESSAGE):
    self.message = message
    super().__init__(self.message)


import json
from time import time
from ..exceptions.sdk_exception import OpenLISAException
from ..exceptions.invalid_command import InvalidCommandException
from ...logging import log

SUCCESS_RESPONSE = "OK"
ERROR_RESPONSE = "ERROR"

COMMAND_DISCONNECT = "DISCONNECT"
COMMAND_GET_INSTRUMENTS = "GET_INSTRUMENTS"
COMMAND_GET_INSTRUMENT = "GET_INSTRUMENT"
COMMAND_GET_INSTRUMENT_COMMANDS = "GET_INSTRUMENT_COMMANDS"
COMMAND_VALIDATE_COMMAND = "VALIDATE_COMMAND"
COMMAND_SEND_COMMAND = "SEND_COMMAND"

class ClientProtocol:
  def __init__(self, message_protocol):
    self._message_protocol = message_protocol

  def __is_valid_response(self, response):
    if response == SUCCESS_RESPONSE:
      return True
    if response == ERROR_RESPONSE:
      return False

    raise Exception("unknown response type: '{}'".format(response))

  def disconnect(self):
    log.debug("[LATENCY_MEASURE][INIT][{}]".format('disconnect'))
    ts = time()
    self._message_protocol.send_msg(COMMAND_GET_INSTRUMENTS)
    te = time()
    log.debug("[LATENCY_MEASURE][FINISH][{}][ELAPSED={} seconds]".format('disconnect', te-ts))
    return

  def get_instruments(self):
    log.debug("[LATENCY_MEASURE][INIT][{}]".format('get_instruments'))
    ts = time()
    self._message_protocol.send_msg(COMMAND_GET_INSTRUMENTS)
    result = self._message_protocol.receive_msg()
    te = time()
    log.debug("[LATENCY_MEASURE][FINISH][{}][ELAPSED={} seconds]".format('get_instruments', te-ts))
    return json.loads(result)

  def get_instrument(self, id):
    log.debug("[LATENCY_MEASURE][INIT][{}]".format('get_instrument'))
    ts = time()
    self._message_protocol.send_msg(COMMAND_GET_INSTRUMENT)
    self._message_protocol.send_msg(id)
    response_type = self._message_protocol.receive_msg()
    result_msg = self._message_protocol.receive_msg()
    te = time()
    log.debug("[LATENCY_MEASURE][FINISH][{}][ELAPSED={} seconds]".format('get_instrument', te-ts))
    if self.__is_valid_response(response_type):
      return json.loads(result_msg)
    else:
      raise OpenLISAException(result_msg)

  def get_instrument_commands(self, id):
    log.debug("[LATENCY_MEASURE][INIT][{}]".format('get_instrument_commands'))
    ts = time()
    self._message_protocol.send_msg(COMMAND_GET_INSTRUMENT_COMMANDS)
    self._message_protocol.send_msg(id)
    response_type = self._message_protocol.receive_msg()
    result_msg = self._message_protocol.receive_msg()
    te = time()
    log.debug("[LATENCY_MEASURE][FINISH][{}][ELAPSED={} seconds]".format('get_instrument_commands', te-ts))
    if self.__is_valid_response(response_type):
      return json.loads(result_msg)
    else:
      raise OpenLISAException(result_msg)

  def validate_command(self, id, command):
    log.debug("[LATENCY_MEASURE][INIT][{}]".format('validate_command'))
    ts = time()
    self._message_protocol.send_msg(COMMAND_VALIDATE_COMMAND)
    self._message_protocol.send_msg(id)
    self._message_protocol.send_msg(command)
    response_type = self._message_protocol.receive_msg()
    if not self.__is_valid_response(response_type):
      err = self._message_protocol.receive_msg()
      te = time()
      log.debug("[LATENCY_MEASURE][FINISH][{}][ELAPSED={} seconds]".format('validate_command', te-ts))
      raise InvalidCommandException("command '{}' is not valid: {}".format(command, err))

    te = time()
    log.debug("[LATENCY_MEASURE][FINISH][{}][ELAPSED={} seconds]".format('validate_command', te-ts))

  def send_command(self, id, command):
    log.debug("[LATENCY_MEASURE][INIT][{}][command={}]".format('send_command', command))
    ts = time()
    self._message_protocol.send_msg(COMMAND_SEND_COMMAND)
    self._message_protocol.send_msg(id)
    self._message_protocol.send_msg(command)
    response_type = self._message_protocol.receive_msg()
    if self.__is_valid_response(response_type):
      format = self._message_protocol.receive_msg()
      response = self._message_protocol.receive_msg(decode=False)
      te = time()
      log.debug("[LATENCY_MEASURE][FINISH][{}][command={}][ELAPSED={} seconds]".format('send_command', command, te-ts))
      return format, response
    else:
      err = self._message_protocol.receive_msg()
      te = time()
      log.debug("[LATENCY_MEASURE][FINISH][{}][command={}][ELAPSED={} seconds]".format('send_command', command, te-ts))
      raise InvalidCommandException("command '{}' for instrument {} is not valid: {}".format(command, id, err))



from time import time
from .domain.exceptions.instrument_not_found import InstrumentNotFoundException
from .logging import log
from .api_client.api_client import ApiClient

DEFAULT_RS232_BAUDRATE = 460800
class SDK:
    def __init__(self, log_level="WARNING", default_string_response_conversion="double", default_bytearray_response_conversion="bytes"):
        log.set_level(log_level)
        log.info("Initializating SDK")
        self._client = ApiClient(str(default_string_response_conversion), str(default_bytearray_response_conversion))

    def connect_through_TCP(self, host, port):
        self._client.connect_through_TCP(host, int(port))

    def connect_through_RS232(self, baudrate=DEFAULT_RS232_BAUDRATE):
        self._client.connect_through_RS232(baudrate)

    def disconnect(self):
        self._client.disconnect()

    def list_instruments(self):
        """
        Returns the list of Instrument objects that are connected and identified by the server
        """
        return self._client.get_instruments()

    def get_instrument(self, id):
        """
        Returns a Instrument object that are connected and identified by the server
        """
        instruments = self.list_instruments()
        for i in instruments:
            if i.ID == id:
                return i

        raise InstrumentNotFoundException




from setuptools import setup, find_packages

VERSION_PLACEHOLDER = '0.3.15'

setup(
  name = 'Open_LISA_SDK',         # How you named your package folder (MyLib)
  packages=find_packages(),
  version = VERSION_PLACEHOLDER,      # Start with a small number and increase it with every change you make
  license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'SDK for Laboratory Instrument Station Adapter',   # Give a short description about your library
  author = 'Ariel Alvarez Windey & Gabriel Robles',                   # Type in your name
  author_email = 'ajalvarez@fi.uba.ar',      # Type in your E-Mail
  url = 'https://github.com/aalvarezwindey/Open-LISA-SDK',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/aalvarezwindey/Open-LISA-SDK/archive/refs/tags/{}.tar.gz'.format(VERSION_PLACEHOLDER),
  keywords = ['SDK', 'ELECTRONIC', 'INSTRUMENT', 'ADAPTER', 'FIUBA', 'OPEN', 'LISA', 'LABORATORY'],   # Keywords that define your package best
  install_requires=["pyserial"],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3.6',      #Specify which pyhton versions that you want to support
  ],
)




import struct

from Open_LISA_SDK.logging import log
from ..exceptions.sdk_exception import OpenLISAException
from ..exceptions.invalid_command import InvalidCommandException

FORMAT_STRING = "str"
FORMAT_DOUBLE = "double"
FORMAT_BYTEARRAY = "bytearray"
FORMAT_BYTES = "bytes"
FORMAT_INT32 = "int"

AVAILABLE_RESPONSE_FORMATS_FOR_BYTEARRAY = [FORMAT_BYTEARRAY, FORMAT_BYTES]
AVAILABLE_RESPONSE_FORMATS_FOR_STRING = [FORMAT_STRING, FORMAT_DOUBLE, FORMAT_INT32]


class Instrument:
  def __init__(self, id, description, brand, model, status, client_protocol, default_string_response_conversion, default_bytearray_response_conversion) -> None:
      self.ID = id
      self.description = description
      self.brand = brand
      self.model = model
      self.status = status
      self._client_protocol = client_protocol
      self._default_string_response_conversion = default_string_response_conversion
      self._default_bytearray_response_conversion = default_bytearray_response_conversion


  def __repr__(self) -> str:
      return 'ID: {}, Description: {}, Brand: {}, Model: {}, Status: {}'.format(self.ID, self.description, self.brand, self.model, self.status)

  def set_default_string_response_conversion(self, v):
    self._default_string_response_conversion = v

  def set_default_bytearray_response_conversion(self, v):
    self._default_bytearray_response_conversion = v

  @staticmethod
  def from_dict(d, client_protocol, def_str, def_bytearr):
    mandatory_keys = ['id', 'description', 'brand', 'model', 'status']
    if all(key in d for key in mandatory_keys):
      return Instrument(id=d['id'], description=d['description'], brand=d['brand'], model=d['model'], status=d['status'], client_protocol=client_protocol, default_string_response_conversion=def_str, default_bytearray_response_conversion=def_bytearr)
    else:
      raise OpenLISAException("missing mandatory keys ({}) in dict {}".format(mandatory_keys, d))

  def available_commands(self):
    # todo: quizas parsear el JSON aca
    return self._client_protocol.get_instrument_commands(self.ID)

  def validate_command(self, command):
    try:
      self._client_protocol.validate_command(self.ID, command)
      print("{} is OK".format(command))
    except InvalidCommandException as e:
      print(e)

  def format_string_response(self, response, format):
    try:
      if format == FORMAT_STRING:
        return response.decode()
      if format == FORMAT_INT32:
        return int(float(response.decode()))
      if format == FORMAT_DOUBLE:
        return float(response.decode())
    except ValueError as e:
      error = "could not convert '{}' to type '{}'.".format(response, format)
      log.error(error)
      raise InvalidCommandException(error)
    except Exception as e:
      log.error(e.message)
      raise InvalidCommandException(e.message)

  def format_bytearray_response(self, response, format):
    try:
      if format == FORMAT_BYTEARRAY:
        return bytearray(response)
      elif format == FORMAT_BYTES:
        return bytes(response)
    except ValueError as e:
      error = "could not convert bytearray response to type '{}'.".format(format)
      log.error(error)
      raise InvalidCommandException(error)
    except Exception as e:
      log.error(e.message)
      raise InvalidCommandException(e.message)

  def send(self, command, convert_to=None):
    format, response = self._client_protocol.send_command(self.ID, command)
    if format == FORMAT_STRING:
      if convert_to:
        if convert_to not in AVAILABLE_RESPONSE_FORMATS_FOR_STRING:
          error = "command '{}' ask for invalid response format type '{}'. Available formats are {}".format(command, format, AVAILABLE_RESPONSE_FORMATS_FOR_BYTEARRAY)
          log.error(error)
          raise InvalidCommandException(error)
        else:
          return self.format_string_response(response, convert_to)
      else:
        return self.format_string_response(response, self._default_string_response_conversion)
    if format == FORMAT_BYTES:
      if convert_to:
        if convert_to not in AVAILABLE_RESPONSE_FORMATS_FOR_BYTEARRAY:
          error = "command '{}' ask for invalid response format type '{}'. Available formats are {}".format(command, format, AVAILABLE_RESPONSE_FORMATS_FOR_BYTEARRAY)
          log.error(error)
          raise InvalidCommandException(error)
        else:
          return self.format_bytearray_response(response, convert_to)
      else:
        return self.format_bytearray_response(response, self._default_bytearray_response_conversion)
